from flask.ext.jsonpify import jsonify

SERVER_OVERLOADED = 503
NOT_FOUND = 404
TIMEOUT = 504
TOO_MANY_REQUESTS = 429


def no_content():
    return "", 204


def error(status, code, message):
    return jsonify({'error': {'code': code, 'message': message}}), status


def not_implemented(message=None):
    return error(501, 'not implemented',
                 message or "Route not implemented yet.")


def model_info_lookup_error(exc):
    return bad_request("Model information could not be retrieved for {0}"
                       .format(exc))


def bad_request(message):
    return error(400, 'bad request', message)


def forbidden(message=None):
    return error(403, 'forbidden',
                 message or "This request requires authentication.")


def not_found(message=None):
    return error(NOT_FOUND, 'not found',
                 message or "Nothing found at this location.")


def server_overloaded(message=None):
    return error(SERVER_OVERLOADED, 'server overloaded',
                 message or ("Cannot process your request because the " +
                             "server is overloaded.  Try again in a" +
                             "few minutes."))


def unknown_error(message):
    return error(500, 'internal server error', message)


def timeout_error(message=None):
    return error(TIMEOUT, 'request_timeout',
                 message or ("Cannot process your request because the " +
                             "server timed out."))


def too_many_requests_error(message=None):
    return error(TOO_MANY_REQUESTS, 'too_many_requests',
                 message or ("A limited number of parallel connections per " +
                             "IP is allowed."))



from pytest import raises

from ..lru import LRU
from ..score_cache import ScoreCache


def test_lru():
    lru = LRU(2)

    cache_context = lru.context("foo", "bar")

    cache_context.store(1, "foo")
    cache_context.store(2, "bar")
    cache_context.store(3, "baz")  # 1 gets bumped

    assert cache_context.lookup(2) == "bar"  # Moves 2 to the front
    cache_context.store(4, "fez")  # 3 gets bumped

    assert cache_context.lookup(2) == "bar"  # Moves 2 to the front
    assert cache_context.lookup(4) == "fez"  # Moves 4 to the front


def test_lru_cache():
    lru = LRU(2)

    cache_context = lru.context("foo", "bar")

    cache_context.store(1, "foo")
    cache_context.store(1, "cachedfoo", injection_cache={"cachedvalue": 1})

    assert cache_context.lookup(1, injection_cache={"cachedvalue": 1}) == \
        "cachedfoo"
    assert cache_context.lookup(1) == "foo"

    with raises(KeyError):
        cache_context.lookup(1, injection_cache={"cachedvalue": 2})


def test_from_config():
    config = {
        'score_caches': {
            'mycache': {
                'class': 'ores.score_caches.LRU',
                'size': 128
            }
        }
    }

    cache = ScoreCache.from_config(config, "mycache")
    context = cache.context("foo", "bar")
    context.store(1, "foo")
    assert context.lookup(1) == "foo"



"""
Class to implement PoolCounter lock manager.

This is useful to make sure too many connections are not coming from one IP.
"""

import logging
import socket
import hashlib

from .lock_manager import LockManager
from ..errors import TimeoutError, TooManyRequestsError

logger = logging.getLogger(__name__)


class PoolCounter(LockManager):
    def __init__(self, nodes):
        self.nodes = nodes
        self.stream = None

    def connect(self, key):
        hashes = []
        for node in self.nodes:
            hashes.append(
                (node, hashlib.sha256(bytes(node[0] + key, 'utf-8')).hexdigest())
            )
        node = sorted(hashes, key=lambda i: i[1])[0][0]
        try:
            self.stream = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
            self.stream.connect(node)
        except ConnectionRefusedError:
            logger.warning(
                'Can not connect to the PoolCounter node %s' % node[0])
            return False
        return True

    def lock(self, key, workers, maxqueue, timeout):
        if not self.stream:
            connected = self.connect(key)
            if not connected:
                return False
        try:
            self.stream.send(bytes(
                'ACQ4ME %s %d %d %d\n' % (key, workers, maxqueue, timeout),
                'utf-8'))
            data = self.stream.recv(4096).decode('utf-8')
        except socket.error as e:
            self.close()
            raise e

        if data.strip() == 'QUEUE_FULL':
            raise TooManyRequestsError

        if data.strip() == 'TIMEOUT':
            raise TimeoutError

        return data.strip() == 'LOCKED'

    def release(self, key):
        if not self.stream:
            return False

        try:
            self.stream.send(bytes('RELEASE %s\n' % key, 'utf-8'))
            data = self.stream.recv(4096).decode('utf-8')
        except socket.error as e:
            self.close()
            raise e

        return data.strip() == 'RELEASED'

    def close(self):
        if self.stream:
            self.stream.close()
            self.stream = None
            return True

        return False

    @classmethod
    def from_config(cls, config, name, section_key="lock_managers"):
        nodes = []
        for node in config[section_key][name]:
            nodes.append((node.split(':')[0],
                          int(node.split(':')[1])))

        return cls(nodes)




class ScoreProcessorOverloaded(RuntimeError):
    pass


class TimeoutError(RuntimeError):
    pass


class MissingContext(RuntimeError):
    pass


class MissingModels(RuntimeError):
    pass


class TooManyRequestsError(RuntimeError):
    pass



from flask import request

from . import util
from ... import preprocessors, responses
from ...util import build_score_request


def configure(config, bp, scoring_system):

    # /v3/scores/
    @bp.route("/v3/scores/", methods=["GET"])
    @preprocessors.nocache
    @preprocessors.minifiable
    def scores_v3():
        try:
            score_request = build_score_request(scoring_system, request)
        except Exception as e:
            return responses.bad_request(str(e))

        return util.build_v3_context_model_map(score_request, scoring_system)

    # /v3/scores/enwiki/?models=reverted&revids=456789|4567890
    @bp.route("/v3/scores/<context>/", methods=["GET"])
    @preprocessors.nocache
    @preprocessors.minifiable
    def score_model_revisions_v3(context):
        try:
            score_request = build_score_request(
                scoring_system, request, context)
        except Exception as e:
            return responses.bad_request(str(e))

        return util.process_score_request(score_request, scoring_system)

    # /v3/scores/enwiki/reverted/?revids=456789|4567890
    @bp.route("/v3/scores/<context>/<int:revid>/", methods=["GET"])
    @preprocessors.nocache
    @preprocessors.minifiable
    def score_revisions_v3(context, revid):
        try:
            score_request = build_score_request(
                scoring_system, request, context, rev_id=revid)
        except Exception as e:
            return responses.bad_request(str(e))

        return util.process_score_request(score_request, scoring_system)

    # /v3/scores/enwiki/reverted/4567890
    @bp.route("/v3/scores/<context>/<int:rev_id>/<model>/", methods=["GET", "POST"])
    @preprocessors.nocache
    @preprocessors.minifiable
    def score_revision_v3(context, model, rev_id):
        try:
            score_request = build_score_request(
                scoring_system, request, context, rev_id=rev_id,
                model_name=model)
        except Exception as e:
            return responses.bad_request(str(e))

        return util.process_score_request(score_request, scoring_system)

    return bp



from pytest import raises

from ..empty import Empty


def test_empty():

    empty = Empty()
    context = empty.context("foo", "bar", version="nope")
    context.store(1, "foo")


def test_key_error():
    with raises(KeyError):
        empty = Empty()
        context = empty.context("foo", "bar", version="nope")
        context.store(1, "foo")
        context.lookup(1)



import glob
import os.path
import yamlconf
import logging
import logging.config
import sys

DEFAULT_DIRS = ["config/", "/etc/ores/"]
DEFAULT_LOGGING_CONFIG = "logging_config.yaml"
DEFAULT_FORMAT = "%(asctime)s %(levelname)s:%(name)s -- %(message)s"


logger = logging.getLogger(__name__)


def build_config(config_dirs=DEFAULT_DIRS, **kwargs):
    # Loads files in alphabetical order based on the bare filename
    config_file_paths = []
    for directory in config_dirs:
        dir_glob = os.path.join(directory, "*.yaml")
        config_file_paths.extend((os.path.basename(path), path)
                                 for path in glob.glob(dir_glob))
    config_file_paths.sort()
    logger.info("Loading configs from {0}".format(config_file_paths))
    config = yamlconf.load(*(open(p) for fn, p in config_file_paths))

    # Add nltk data path if specified.
    if 'data_paths' in config['ores'] and \
       'nltk' in config['ores']['data_paths']:
        import nltk
        nltk.data.path.append(config['ores']['data_paths']['nltk'])

    return config


def configure_logging(verbose=False, debug=False, logging_config=None, **kwargs):
    # Load logging config if specified.  If no config file is specified, we
    # make a half-hearted attempt to find a distributed logging_config.yaml
    # in the current working directory.
    if logging_config is None:
        if os.path.exists(DEFAULT_LOGGING_CONFIG):
            logging_config = DEFAULT_LOGGING_CONFIG

    if logging_config is not None:
        with open(logging_config) as f:
            logging_config = yamlconf.load(f)
            logging.config.dictConfig(logging_config)

        # Secret sauce: if running from the console, mirror logs there.
        if sys.stdin.isatty():
            handler = logging.StreamHandler(stream=sys.stderr)
            formatter = logging.Formatter(fmt=DEFAULT_FORMAT)
            handler.setFormatter(formatter)
            logging.getLogger().addHandler(handler)

    else:
        # Configure fallback logging.
        logging.basicConfig(level=logging.INFO, format=DEFAULT_FORMAT)
        logging.getLogger('requests').setLevel(logging.INFO)
        logging.getLogger('stopit').setLevel(logging.ERROR)

    # Command-line options can override some of the logging config, regardless
    # of whether logging_config.yaml was provided.
    # TODO: Document and reconcile debug vs verbose.
    if debug:
        logging.getLogger().setLevel(logging.DEBUG)
        logging.getLogger('ores.metrics_collectors.logger') \
               .setLevel(logging.DEBUG)

    if verbose:
        logging.getLogger('revscoring.dependencies.dependent') \
               .setLevel(logging.DEBUG)



"""
Runs a basic test against an ORES api assuming that it supports the standard
configuration "/v2/testwiki/revid/..."

:Usage:
    test_api -h | --help
    test_api <ores-url> [--debug]

:Options:
    -h --help        Prints this documentation
    <ores-url>       URL of base ORES webserver
    --debug          Print debugging information
"""
import json
import logging

import docopt
import requests

logger = logging.getLogger(__name__)


def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)

    logging.basicConfig(
        level=logging.INFO if not args['--debug'] else logging.DEBUG,
        format='%(asctime)s %(levelname)s:%(name)s -- %(message)s'
    )
    # Requests is loud.  Be quiet requests.
    requests.packages.urllib3.disable_warnings()

    ores_url = args['<ores-url>']
    make_request(ores_url, "/")
    make_request(ores_url, "/ui/")
    make_request(ores_url, "/scores/", is_json=True)
    make_request(ores_url, "/v1/spec/", is_json=True)
    make_request(ores_url, "/v2/spec/", is_json=True)
    make_request(ores_url, "/v3/spec/", is_json=True)
    make_request(
        ores_url, "/v1/scores/testwiki/revid/2342342/", is_json=True,
        equal_to={"2342342": {"prediction": False,
                              "probability": {"false": 0.76, "true": 0.24}}})
    make_request(
        ores_url, "/v2/scores/testwiki/revid/2342342/", is_json=True,
        equal_to={"scores": {"testwiki": {"revid": {"scores": {
            "2342342": {
                "prediction": False,
                "probability": {"false": 0.76, "true": 0.24}
            }
        }, "version": "0.0.0"}}}})
    make_request(
        ores_url, "/v3/scores/testwiki/2342342/revid/", is_json=True,
        equal_to={"testwiki": {
            "models": {"revid": {"version": "0.0.0"}},
            "scores": {"2342342": {
                "revid": {"score": {"prediction": False,
                                    "probability": {"false": 0.76,
                                                    "true": 0.24}}}
            }}}})
    response = requests.get(ores_url + "/404/")
    assert response.status_code == 404, "/404/ didn't get a 404!"

    make_request(
        ores_url,
        "/v3/scores/testwiki/2342342/revid/?features&feature.delay=16",
        is_json=True,
        equal_to={"testwiki": {
            "models": {"revid": {"version": "0.0.0"}},
            "scores": {"2342342": {
                "revid": {"error": {'message': 'Timed out after 15 seconds.',
                                    'type': 'TimeoutError'}}
            }}}})

    other_wiki_event = {
        "comment": "/* K-O */", "database": "enwiki",
        "meta": {
            "domain": "en.wikipedia.org", "dt": "2017-12-05T15:56:51+00:00",
            "id": "e87a7723-d9d4-11e7-9e8e-141877613bad",
            "request_id": "3464552a-85d0-404e-aa24-80b74473b15f",
            "schema_uri": "mediawiki/revision/create/2",
            "topic": "eqiad.mediawiki.revision-create",
            "uri": "https://en.wikipedia.org/wiki/List_of_Ateneo_de_Manila_University_people",
            "partition": 0, "offset": 561544941
        },
        "page_id": 4716305, "page_is_redirect": False,
        "page_namespace": 0,
        "page_title": "List_of_Ateneo_de_Manila_University_people",
        "parsedcomment": "<a href=\"/wiki/List_of_Ateneo_de_Manila_University_" +
                         "people#K-O\" title=\"List of Ateneo de Manila " +
                         "University people\">→</a>‎<span dir=\"auto\">" +
                         "<span class=\"autocomment\">K-O</span></span>",
        "performer": {
            "user_edit_count": 13845,
            "user_groups": ["extendedconfirmed", "*", "user", "autoconfirmed"],
            "user_id": 24365224, "user_is_bot": False,
            "user_registration_dt": "2015-03-09T02:40:31Z",
            "user_text": "Khendygirl"
        },
        "rev_content_changed": True, "rev_content_format": "wikitext",
        "rev_content_model": "wikitext", "rev_id": 813852458,
        "rev_len": 60647, "rev_minor_edit": False,
        "rev_parent_id": 813852231,
        "rev_sha1": "0a6fggdfff46x0bptqle6pycuz0paht",
        "rev_timestamp": "2017-12-05T15:56:51Z"
    }
    make_request(
        ores_url,
        "/v3/precache",
        post_json=other_wiki_event,
        http_code=204)

    own_event = other_wiki_event
    own_event['database'] = "testwiki"
    make_request(
        ores_url,
        "/v3/precache",
        is_json=True,
        post_json=own_event,
        http_code=200,
        equal_to={
            "testwiki": {
                "models": {"revid": {"version": "0.0.0"}},
                "scores": {"813852458": {"revid": {"score": {
                                "prediction": True,
                                "probability": {
                                    "false": 0.15000000000000002,
                                    "true": 0.85
                                }}}}}
            }})


def make_request(ores_url, path, http_code=200, is_json=False,
                 equal_to=None, post_json=None):
    logger.debug("Requesting {0}".format(path))
    if post_json is None:
        response = requests.get(ores_url + path)
    else:
        response = requests.post(ores_url + path, json=post_json)
    assert response.status_code == http_code, \
           "Status code mismatch {0} for {1}: {2}".format(
               response.status_code, path, response.content)

    content = response.text

    if is_json:
        try:
            content = json.loads(content)
        except ValueError:
            raise RuntimeError("Could not parse the following as JSON: '{0}"
                               .format(content[0:100]))

    if equal_to:
        assert content == equal_to, "{0} != {1}".format(content, equal_to)

    return response



import time

from revscoring import Model
from revscoring.datasources.revision_oriented import revision
from revscoring.features import Feature
from revscoring.scoring import ModelInfo
from revscoring.scoring.statistics import Classification


def process_reversed_last_two_in_rev_id(rev_id):
    last_two = str(rev_id)[-2:]
    if len(last_two) == 1:
        return int(last_two + "0")
    else:
        return int("".join(reversed(last_two)))


reversed_last_two_in_rev_id = Feature(
    "revision.reversed_last_two_in_rev_id",
    process_reversed_last_two_in_rev_id,
    returns=int,
    depends_on=[revision.id]
)


def process_delay():
    return 0.0


delay = Feature("delay", process_delay, returns=float)


class RevIdScorer(Model):
    """
    Implements a basic, testing scorer that predicts whether a revision ID's
    reversed last two digits are greater than 50.

    E.g. 974623 = 32 and 23754929 = 92
    """
    def __init__(self, version=None):
        super().__init__([reversed_last_two_in_rev_id, delay], version=version)
        self.info = ModelInfo()
        self.info['version'] = version
        self.info['type'] = "RevIDScorer"
        self.info['behavior'] = "Returns the last two digits in a rev_id " + \
                                "as a score."
        self.info['statistics'] = self.calculate_statistics()

    def score(self, feature_values):
        last_two_in_rev_id, delay = feature_values
        time.sleep(delay)
        probability = last_two_in_rev_id / 100

        if probability > 0.5:
            prediction = True
        else:
            prediction = False

        return {
            'prediction': prediction,
            'probability': {
                True: probability,
                False: 1 - probability
            }
        }

    def calculate_statistics(self):
        "Jam some data through to generate statistics"
        rev_ids = range(0, 100, 1)
        feature_values = zip(rev_ids, [0] * 100)
        scores = [self.score(f) for f in feature_values]
        labels = [s['prediction'] for s in scores]
        statistics = Classification(labels, threshold_ndigits=1, decision_key='probability')
        score_labels = list(zip(scores, labels))
        statistics.fit(score_labels)
        return statistics

    @classmethod
    def from_config(cls, config, name, section_key='scorer_models'):
        section = config[section_key][name]
        return cls(**{k: v for k, v in section.items() if k != "class"})



"""
Runs a celery worker process that exposes a task API for
ores.scoring_systems.CeleryQueue

Usage:
    celery [--config-dir=<path>]... [--logging-config=<path>]
           [--debug] [--verbose]

Options:
    -h --help                Prints this documentation
    --config-dir=<path>      The path to a directory containing configuration
                             [default: config/]
    --logging-config=<path>  The path to a logging configuration file
    --debug                  Print debug logging information
    --verbose                Print verbose extraction information
"""
import docopt

from ..scoring_systems import CeleryQueue
from .util import build_config, configure_logging


def main(argv=None):
    args = docopt.docopt(__doc__, argv=argv)
    verbose = args['--verbose']
    debug = args['--debug']

    run(verbose=verbose, debug=debug,
        logging_config=args['--logging-config'],
        config_dirs=args['--config-dir'])


def run(**kwargs):
    application = build(**kwargs)
    application.worker_main(
        argv=["celery_worker"])


def build(**kwargs):
    configure_logging(**kwargs)

    config = build_config(**kwargs)
    scoring_system = CeleryQueue.from_config(
        config, config['ores']['scoring_system'])
    return scoring_system.application



from ..score_request import ScoreRequest


def test_score_request():
    sr = ScoreRequest("foo", [1, 2, 3], ["bar", "baz"])

    assert sr.context_name == "foo"
    assert sr.rev_ids == {1, 2, 3}
    assert sr.model_names == {"bar", "baz"}
    assert sr.precache is False
    assert sr.include_features is False
    assert sr.model_info is None



import traceback

from flask import request
from revscoring.errors import ModelInfoLookupError

from . import util
from ... import preprocessors, responses
from .... import errors
from ...util import build_score_request, jsonify


def configure(config, bp, scoring_system):

    # /scores/
    @bp.route("/scores/", methods=["GET"])
    @bp.route("/v1/scores/", methods=["GET"])
    @preprocessors.nocache
    @preprocessors.minifiable
    def scores():
        contexts = [context for context in scoring_system]

        contexts.sort()

        return jsonify({'contexts': contexts})

    def process_score_request(score_request, context=None, model=None):
        try:
            if len(score_request.rev_ids) == 0:
                return util.format_some_model_info(
                    scoring_system, score_request, model)
            else:
                score_response = scoring_system.score(score_request)
                return util.format_v1_score_response(score_response, model)
        except errors.ScoreProcessorOverloaded:
            scoring_system.metrics_collector.response_made(
                responses.SERVER_OVERLOADED, score_request)
            return responses.server_overloaded()
        except errors.MissingContext as e:
            scoring_system.metrics_collector.response_made(
                responses.NOT_FOUND, score_request)
            return responses.not_found("No scorers available for {0}"
                                       .format(e))
        except errors.MissingModels as e:
            scoring_system.metrics_collector.response_made(
                responses.NOT_FOUND, score_request)
            context_name, model_names = e.args
            return responses.not_found(
                "Models {0} not available for {1}"
                .format(tuple(model_names), context_name))
        except ModelInfoLookupError as e:
            return responses.model_info_lookup_error(e)
        except errors.TimeoutError:
            scoring_system.metrics_collector.response_made(
                responses.TIMEOUT, score_request)
            return responses.timeout_error()
        except errors.TooManyRequestsError:
            scoring_system.metrics_collector.response_made(
                responses.TOO_MANY_REQUESTS, score_request)
            return responses.too_many_requests_error()
        except Exception:
            return responses.unknown_error(traceback.format_exc())

    # /scores/enwiki/?models=reverted&revids=456789|4567890
    @bp.route("/scores/<context>/", methods=["GET"])
    @bp.route("/v1/scores/<context>/", methods=["GET"])
    @preprocessors.nocache
    @preprocessors.minifiable
    def score_model_revisions(context):
        try:
            score_request = build_score_request(
                scoring_system, request, context)
        except Exception as e:
            return responses.bad_request(str(e))
        return process_score_request(score_request, context)

    # /scores/enwiki/reverted/?revids=456789|4567890
    @bp.route("/scores/<context>/<model>/", methods=["GET"])
    @bp.route("/v1/scores/<context>/<model>/", methods=["GET"])
    @preprocessors.nocache
    @preprocessors.minifiable
    def score_revisions(context, model):
        try:
            score_request = build_score_request(
                scoring_system, request, context, model_name=model)
        except Exception as e:
            return responses.bad_request(str(e))

        return process_score_request(score_request, context, model)

    # /scores/enwiki/reverted/4567890
    @bp.route("/scores/<context>/<model>/<int:rev_id>/", methods=["GET", "POST"])
    @bp.route("/v1/scores/<context>/<model>/<int:rev_id>/", methods=["GET", "POST"])
    @preprocessors.nocache
    @preprocessors.minifiable
    def score_revision(context, model, rev_id):
        try:
            score_request = build_score_request(
                scoring_system, request, context, rev_id=rev_id,
                model_name=model)
        except Exception as e:
            return responses.bad_request(str(e))

        return process_score_request(score_request, context, model)

    return bp



import sys
import platform
from pkg_resources import VersionConflict

from .about import (__author__, __author_email__, __description__, __name__,
                    __url__, __version__)

if sys.version_info <= (3, 0):
    raise VersionConflict("ORES requires Python '>=3' but your Python version is " + platform.python_version())


__all__ = [__name__, __version__, __author__, __author_email__,
           __description__, __url__]



"""Lock manager interface."""


class LockManager:
    def connect(self):
        """
        Connect to the lock manager.

        returns false if it wasn't able to connect, true otherwise.
        """
        raise NotImplementedError

    def lock(self, key, workers, maxqueue, timeout):
        """
        Acquire a lock in the lock manager.

        returns false if it wasn't able to lock, true otherwise.
        """
        raise NotImplementedError

    def release(self, key):
        """
        Release the lock in the lock manager.

        returns false if it wasn't able to release, true otherwise.
        """
        raise NotImplementedError

    def close(self):
        """
        Close the connection to the lock manager.

        returns false if it wasn't able to close the connection, true otherwise.
        """
        raise NotImplementedError



from ..rev_id_scorer import RevIdScorer


def test_statistics():
    scorer = RevIdScorer()
    stats = scorer.info['statistics']

    # Debug output if the test crashes.
    print(stats.format_str({}, threshold_ndigits=1))
    print(stats['thresholds'].format_str({}, threshold_ndigits=1))

    assert round(stats['roc_auc']['micro'], 3) == 0.996
    assert round(stats['pr_auc']['micro'], 3) == 0.992

    optimized = stats.lookup(
        '"maximum recall @ !precision >= 0.8".labels.true')
    assert optimized == 0.5
